@@ -11,15 +11,19 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+from typing import Any, Optional, Sequence, Dict
 
 import zlib
 import io
 import numpy as np
 
+import pyspng
 import simplejpeg
 import compresso
 import fastremap
 
+from PIL import Image
+
 from .lib import yellow
 
 try:
@@ -42,7 +46,17 @@ except ImportError:
     def decompress(cls, content):
       raise NotImplementedError(fpziperrormsg)
 
-def encode(img_chunk, encoding, block_size=None):
+SUPPORTED_ENCODINGS = (
+  "raw", "kempressed", "fpzip",
+  "compressed_segmentation", "compresso",
+  "jpeg", "png"
+)
+
+def encode(
+  img_chunk:np.ndarray, 
+  encoding:str, 
+  block_size:Optional[Sequence[int]] = None
+) -> bytes:
   if encoding == "raw":
     return encode_raw(img_chunk)
   elif encoding == "kempressed":
@@ -56,6 +70,8 @@ def encode(img_chunk, encoding, block_size=None):
     return compresso.compress(img_chunk[:,:,:,0])
   elif encoding == "jpeg":
     return encode_jpeg(img_chunk)
+  elif encoding == "png":
+    return encode_png(img_chunk)
   elif encoding == "npz":
     return encode_npz(img_chunk)
   elif encoding == "npz_uint8":
@@ -66,10 +82,13 @@ def encode(img_chunk, encoding, block_size=None):
     raise NotImplementedError(encoding)
 
 def decode(
-  filedata, encoding, 
-  shape=None, dtype=None, 
-  block_size=None, background_color=0
-):
+  filedata:bytes, 
+  encoding:str, 
+  shape:Optional[Sequence[int]] = None, 
+  dtype:Any = None, 
+  block_size:Optional[Sequence[int]] = None, 
+  background_color:int = 0
+) -> np.ndarray:
   if (shape is None or dtype is None) and encoding not in ('npz', 'fpzip', 'kempressed'):
     raise ValueError("Only npz encoding can omit shape and dtype arguments. {}".format(encoding))
 
@@ -87,15 +106,14 @@ def decode(
     return compresso.decompress(filedata).reshape(shape)
   elif encoding == "jpeg":
     return decode_jpeg(filedata, shape=shape, dtype=dtype)
+  elif encoding == "png":
+    return decode_png(filedata, shape=shape, dtype=dtype)
   elif encoding == "npz":
     return decode_npz(filedata)
   else:
     raise NotImplementedError(encoding)
 
-def encode_jpeg(arr, quality=85):
-  if not np.issubdtype(arr.dtype, np.uint8):
-    raise ValueError("Only accepts uint8 arrays. Got: " + str(arr.dtype))
-
+def as2d(arr):
   # simulate multi-channel array for single channel arrays
   while arr.ndim < 4:
     arr = arr[..., np.newaxis] # add channels to end of x,y,z
@@ -106,21 +124,37 @@ def encode_jpeg(arr, quality=85):
   reshaped = reshaped.reshape(
     reshaped.shape[0] * reshaped.shape[1], reshaped.shape[2], num_channel
   )
+  return reshaped, num_channel
+
+def encode_jpeg(arr, quality=85):
+  if not np.issubdtype(arr.dtype, np.uint8):
+    raise ValueError("Only accepts uint8 arrays. Got: " + str(arr.dtype))
+
+  arr, num_channel = as2d(arr)
+  arr = np.ascontiguousarray(arr)
+
   if num_channel == 1:
     return simplejpeg.encode_jpeg(
-      np.ascontiguousarray(reshaped), 
+      arr, 
       colorspace="GRAY",
       colorsubsampling="GRAY",
       quality=quality,
     )
   elif num_channel == 3:
     return simplejpeg.encode_jpeg(
-      np.ascontiguousarray(reshaped),
+      arr,
       colorspace="RGB",
       quality=quality,
     )
   raise ValueError("Number of image channels should be 1 or 3. Got: {}".format(arr.shape[3]))
 
+def encode_png(arr, compress_level=9):
+  if arr.dtype not in (np.uint8, np.uint16):
+    raise ValueError("Only accepts uint8 and uint16 arrays. Got: " + str(arr.dtype))
+
+  arr, num_channel = as2d(arr)
+  return pyspng.encode(arr, compress_level=compress_level)
+
 def encode_npz(subvol):
   """
   This file format is unrelated to np.savez
@@ -137,7 +171,11 @@ def encode_npz(subvol):
   cdz = zlib.compress(fileobj.getvalue())
   return cdz
 
-def encode_compressed_segmentation(subvol, block_size, accelerated=ACCELERATED_CSEG):
+def encode_compressed_segmentation(
+  subvol:np.ndarray, 
+  block_size:Sequence[int], 
+  accelerated:bool = ACCELERATED_CSEG
+) -> bytes:
   assert np.dtype(subvol.dtype) in (np.uint32, np.uint64)
 
   if accelerated:
@@ -189,6 +227,11 @@ def decode_jpeg(bytestring, shape, dtype):
   ).ravel()
   return data.reshape(shape, order='F')
 
+def decode_png(bytestring: bytes, shape, dtype):
+  img = pyspng.load(bytestring).reshape(-1)
+  img = img.astype(dtype, copy=False)
+  return img.reshape(shape, order='F')
+
 def decode_raw(bytestring, shape, dtype):
   return np.frombuffer(bytearray(bytestring), dtype=dtype).reshape(shape, order='F')
 
@@ -210,10 +253,10 @@ def decode_compressed_segmentation_pure_python(bytestring, shape, dtype, block_s
   return chunk.T
 
 def labels(
-  filedata, encoding, 
+  filedata:bytes, encoding:str, 
   shape=None, dtype=None, 
-  block_size=None, background_color=0
-):
+  block_size=None, background_color:int = 0
+) -> np.ndarray:
   """
   Extract unique labels from a chunk using
   the most efficient means possible for the
@@ -236,4 +279,55 @@ def labels(
   else:
     raise NotImplementedError(f"Encoding {encoding} is not supported. Try: raw, compressed_segmentation, or compresso.")
 
+def remap(
+  filedata:bytes, encoding:str, 
+  mapping:Dict[int,int],
+  preserve_missing_labels=False,
+  shape=None, dtype=None,
+  block_size=None
+) -> bytes:
+  if filedata is None or len(filedata) == 0:
+    return filedata
+  elif encoding == "compressed_segmentation":
+    return cseg.remap(
+      filedata, shape, dtype, mapping, 
+      preserve_missing_labels=preserve_missing_labels, block_size=block_size
+    )
+  elif encoding == "compresso":
+    return compresso.remap(filedata, mapping, preserve_missing_labels=preserve_missing_labels)
+  else:
+    img = decode(filedata, encoding, shape, dtype, block_size)
+    fastremap.remap(img, mapping, preserve_missing_labels=preserve_missing_labels, in_place=True)
+    return encode(img, encoding, block_size)
+
+def read_voxel(
+  xyz:Sequence[int], 
+  filedata:bytes, 
+  encoding:str, 
+  shape:Optional[Sequence[int]] = None, 
+  dtype:Any = None, 
+  block_size:Optional[Sequence[int]] = None,
+  background_color:int = 0
+) -> np.ndarray:
+  if encoding == "compressed_segmentation":
+    arr = cseg.CompressedSegmentationArray(
+      filedata, shape=shape[:3], dtype=dtype, block_size=block_size
+    )
+    out = np.empty((1,1,1,1), dtype=dtype, order="F")
+    out[0,0,0,0] = arr[tuple(xyz)]
+    return out
+  elif encoding == "compresso":
+    arr = compresso.CompressoArray(filedata)
+    out = np.empty((1,1,1,1), dtype=dtype, order="F")
+    out[0,0,0,0] = arr[tuple(xyz)]
+    return out
+  else:
+    img = decode(filedata, encoding, shape, dtype, block_size, background_color)
+    return img[tuple(xyz)][:, np.newaxis, np.newaxis, np.newaxis]
+
+
+
+
+
+
 
